
#include "Variables_lib.h"

using namespace libflowstar;

Variables::Variables()
{
}

Variables::~Variables()
{
	varTab.clear();
	varNames.clear();
}

Variables::Variables(const Variables & variables)
{
	varTab		= variables.varTab;
	varNames	= variables.varNames;
}

Variables & Variables::operator = (const Variables & variables)
{
	if(this == &variables)
		return *this;

	varTab		= variables.varTab;
	varNames	= variables.varNames;

	return *this;
}

bool Variables::declareVar(const std::string & vName)
{
	std::map<std::string,int>::const_iterator iter;

	if((iter = varTab.find(vName)) == varTab.end())
	{
		varTab[vName] = varNames.size();
		varNames.push_back(vName);
		return true;
	}
	else
	{
		return false;
	}
}

int Variables::getIDForVar(const std::string & vName) const
{
	std::map<std::string,int>::const_iterator iter;
	if((iter = varTab.find(vName)) == varTab.end())
	{
		return -1;
	}

	return iter->second;
}

bool Variables::getVarName(std::string & vName, const int id) const
{
	if(id >= 0 && id < varNames.size())
	{
		vName = varNames[id];
		return true;
	}
	else
	{
		return false;
	}
}

int Variables::size() const
{
	return varNames.size();
}

void Variables::clear()
{
	varTab.clear();
	varNames.clear();
}

void Variables::output(FILE *fp) const
{
	std::map<std::string,int>::const_iterator iter;

	for(iter=varTab.begin(); iter!=varTab.end(); ++iter)
	{
		fprintf(fp, "ID = %d, Name = %s\n", iter->second, iter->first.c_str());
	}
}

