/*
 * insulin.cpp
 *
 *  Created on: March 10, 2017
 *      Author: xinchen
 */

#include "../flowstar-2.0.0/Continuous_lib.h"

using namespace libflowstar;


class HybridConfig
{
public:
	Flowpipe flowpipe;
	int modeID;
	int steps;
	int jumps;

	HybridConfig();
	HybridConfig(const Flowpipe & fp, const int mode, const int n, const int m)
	{
		flowpipe = fp;
		modeID = mode;
		steps = n;
		jumps = m;
	}

	HybridConfig(const HybridConfig & config)
	{
		flowpipe = config.flowpipe;
		modeID = config.modeID;
		steps = config.steps;
		jumps = config.jumps;
	}

	~HybridConfig() { }

	HybridConfig & operator = (const HybridConfig & config)
	{
		if(this == &config)
			return *this;

		flowpipe = config.flowpipe;
		modeID = config.modeID;
		steps = config.steps;
		jumps = config.jumps;

		return *this;
	}
};


int main()
{
	Variables stateVars;

	// declare the state variables, the first one should always be the local time variable
	stateVars.declareVar("local_t");
	stateVars.declareVar("X");
	stateVars.declareVar("Isc1");
	stateVars.declareVar("Isc2");
	stateVars.declareVar("Gt");
	stateVars.declareVar("Gp");
	stateVars.declareVar("Il");
	stateVars.declareVar("Ip");
	stateVars.declareVar("I1");
	stateVars.declareVar("Id");
	stateVars.declareVar("Gs");
	stateVars.declareVar("t");
	stateVars.declareVar("uc");


	Polynomial deriv_X("-0.0278 * X +0.0278 * ( 18.2129 * Ip - 100.25)", stateVars);
	Polynomial deriv_Isc1("-0.0171 * Isc1", stateVars);
	Polynomial deriv_Isc2("0.0152 * Isc1 - 0.0078 * Isc2", stateVars);
	Polynomial deriv_Gt("-0.0039* (3.2267+0.0313*X) * Gt * ( 1 - 0.0026 * Gt + (2.5097e-6) *Gt^2) + 0.0581 *Gp - 0.0871* Gt", stateVars);
	Polynomial deriv_Gp("4.7314 - 0.0047 * Gp - 0.0121 * Id - 1 - 0.0581*Gp +  0.0871*Gt", stateVars);
	Polynomial deriv_Il("-0.4219 * Il + 0.2250* Ip", stateVars);
	Polynomial deriv_Ip("-0.3150 * Ip + 0.1545 * Il + 0.0019 * Isc1 + 0.0078 * Isc2", stateVars);
	Polynomial deriv_I1("-0.0046 * (I1 - 18.2129 * Ip)", stateVars);
	Polynomial deriv_Id("-0.0046 * (Id - I1 )", stateVars);
	Polynomial deriv_Gs("0.1*(0.5221 * Gp - Gs)", stateVars);
	Polynomial deriv_t("1", stateVars);
	Polynomial deriv_uc("0", stateVars);


	// Insulin Rates Strategy 2
	Polynomial insulin_rate_0("0.97751710655*0.1", stateVars);
	Polynomial insulin_rate_1("0.97751710655*0.3", stateVars);
	Polynomial insulin_rate_2("0.97751710655*0.7", stateVars);
	Polynomial insulin_rate_3("0.97751710655*1.2", stateVars);
	Polynomial insulin_rate_4("0.97751710655*1.5", stateVars);




	// meal coefficients
	Polynomial meal_coeff_1("(1.140850553428184e-4)*t^2*uc + (6.134609247812877e-5)*t*uc", stateVars);
	Polynomial meal_coeff_2("(5.252705445621361e-5)*t^2*uc - 0.007468661585336*t*uc + 0.281302431132355*uc", stateVars);
	Polynomial meal_coeff_3("(1.245404770117318e-7)*t^2*uc - (9.112517786233733e-5)*t*uc + 0.026475608001390*uc", stateVars);
	Polynomial meal_coeff_4("-(6.307316106980570e-5)*t^2*uc + 0.048262107570417*t*uc - 9.190266060911723*uc", stateVars);
	Polynomial meal_coeff_5("(3.552954405332104e-6)*t^2*uc - 0.003423642950746*t*uc + 0.823855671531609*uc", stateVars);
	Polynomial meal_coeff_6("(1.113158675054563e-8)*t^2*uc - (1.482047571340105e-5)*t*uc + 0.004900138660598*uc", stateVars);



	// define the ODE
	ODE ode(stateVars);
	ode.assignDerivative("X", deriv_X);
	ode.assignDerivative("Isc1", deriv_Isc1);
	ode.assignDerivative("Isc2", deriv_Isc2);
	ode.assignDerivative("Gt", deriv_Gt);
	ode.assignDerivative("Gp", deriv_Gp);
	ode.assignDerivative("Il", deriv_Il);
	ode.assignDerivative("Ip", deriv_Ip);
	ode.assignDerivative("I1", deriv_I1);
	ode.assignDerivative("Id", deriv_Id);
	ode.assignDerivative("Gs", deriv_Gs);
	ode.assignDerivative("t", deriv_t);
	ode.assignDerivative("uc", deriv_uc);


	// constraints
	PolynomialConstraint G_geq_80("-Gs + 80", stateVars);
	PolynomialConstraint G_leq_80("Gs - 80", stateVars);

	PolynomialConstraint G_leq_75("Gs - 75", stateVars);
	PolynomialConstraint G_geq_75("-Gs + 75", stateVars);

	PolynomialConstraint G_geq_120("-Gs + 120", stateVars);
	PolynomialConstraint G_leq_120("Gs - 120", stateVars);

	PolynomialConstraint G_leq_115("Gs - 115", stateVars);
	PolynomialConstraint G_geq_115("-Gs + 115", stateVars);

	PolynomialConstraint G_geq_180("-Gs + 180", stateVars);
	PolynomialConstraint G_leq_180("Gs - 180", stateVars);

	PolynomialConstraint G_leq_175("Gs - 175", stateVars);
	PolynomialConstraint G_geq_175("-Gs + 175", stateVars);

	PolynomialConstraint G_geq_300("-Gs + 300", stateVars);
	PolynomialConstraint G_leq_300("Gs - 300", stateVars);

	PolynomialConstraint G_leq_295("Gs - 295", stateVars);
	PolynomialConstraint G_geq_295("-Gs + 295", stateVars);


	// jump conditions
	std::vector<PolynomialConstraint> mode_0_to_1;
	mode_0_to_1.push_back(G_geq_80);

	std::vector<PolynomialConstraint> mode_1_to_0;
	mode_1_to_0.push_back(G_leq_75);

	std::vector<PolynomialConstraint> mode_1_to_2;
	mode_1_to_2.push_back(G_geq_120);

	std::vector<PolynomialConstraint> mode_2_to_1;
	mode_2_to_1.push_back(G_leq_115);

	std::vector<PolynomialConstraint> mode_2_to_3;
	mode_2_to_3.push_back(G_geq_180);

	std::vector<PolynomialConstraint> mode_3_to_2;
	mode_3_to_2.push_back(G_leq_175);

	std::vector<PolynomialConstraint> mode_3_to_4;
	mode_3_to_4.push_back(G_geq_300);

	std::vector<PolynomialConstraint> mode_4_to_3;
	mode_4_to_3.push_back(G_leq_295);


	std::vector<PolynomialConstraint> inv_mode_0;
	inv_mode_0.push_back(G_leq_80);

	std::vector<PolynomialConstraint> inv_mode_1;
	inv_mode_1.push_back(G_geq_75);
	inv_mode_1.push_back(G_leq_120);

	std::vector<PolynomialConstraint> inv_mode_2;
	inv_mode_2.push_back(G_geq_115);
	inv_mode_2.push_back(G_leq_180);

	std::vector<PolynomialConstraint> inv_mode_3;
	inv_mode_3.push_back(G_geq_175);
	inv_mode_3.push_back(G_leq_300);

	std::vector<PolynomialConstraint> inv_mode_4;
	inv_mode_4.push_back(G_geq_295);



	double stepsize = 0.2;

	// reachability setting
	ReachabilitySetting rs;

	rs.setFixedStepsize(stepsize);
	rs.setFixedOrder(3);
	rs.setPrecondition(ID_PRE);

	// cutoff threshold
	Interval cutoff(-1e-8,1e-8);
	rs.setCutoff(cutoff);
	rs.setPrecision(256);

	// remainder estimation
	Interval I(-1e-2,1e-2);

	Assignment<Interval> estimation(13);
	estimation.assign("X", I, stateVars);
	estimation.assign("Isc1", I, stateVars);
	estimation.assign("Isc2", I, stateVars);
	estimation.assign("Gt", I, stateVars);
	estimation.assign("Gp", I, stateVars);
	estimation.assign("Il", I, stateVars);
	estimation.assign("Ip", I, stateVars);
	estimation.assign("I1", I, stateVars);
	estimation.assign("Id", I, stateVars);
	estimation.assign("Gs", I, stateVars);
	estimation.assign("t", I, stateVars);
	estimation.assign("uc", I, stateVars);

	rs.setRemainderEstimation(estimation);
	rs.prepareForReachability();


	Interval init_X(0,0), init_Isc1(72.43, 72.43), init_Isc2(141.15, 141.15), init_Gt(162.45, 162.45),
			init_Il(3.2, 3.2), init_Ip(5.5,5.5), init_I1(100.25,100.25), init_Id(100.25,100.25),
			init_Gs(120,160), init_t(0,0), init_uc(50,90);

	Interval init_Gp = init_Gs;
	init_Gp.mul_assign(1.9152);

	Assignment<Interval> box(13);
	box.assign("X", init_X, stateVars);
	box.assign("Isc1", init_Isc1, stateVars);
	box.assign("Isc2", init_Isc2, stateVars);
	box.assign("Gt", init_Gt, stateVars);
	box.assign("Gp", init_Gp, stateVars);
	box.assign("Il", init_Il, stateVars);
	box.assign("Ip", init_Ip, stateVars);
	box.assign("I1", init_I1, stateVars);
	box.assign("Id", init_Id, stateVars);
	box.assign("Gs", init_Gs, stateVars);
	box.assign("t", init_t, stateVars);
	box.assign("uc", init_uc, stateVars);

	Flowpipe initialSet(box);

	std::list<HybridConfig> queue;

	HybridConfig initialConfig(initialSet, 2, 0, 0);
	queue.push_back(initialConfig);

	std::list<Flowpipe> reachableSet;

	// switch triggered every 5 minutes, 0.2*25 = 5
	int N = 25;
	int total_steps = 720 / stepsize;		// should be an integer
	int max_jumps = 10;


	clock_t begin, end;
	begin = clock();

	for(; queue.size() > 0;)
	{
		// test: print out the queue content
		std::list<HybridConfig>::iterator queue_iter = queue.begin();
		for(; queue_iter != queue.end(); ++queue_iter)
		{
			printf("(%f, %d, %d)\t", queue_iter->steps*stepsize, queue_iter->modeID, queue_iter->jumps);
		}
		printf("\n");


		HybridConfig current_config = queue.front();
		queue.pop_front();

		switch(current_config.modeID)
		{
		case 0:
		{
			Polynomial deriv = deriv_Isc1 + insulin_rate_0;
			ode.assignDerivative("Isc1", deriv);
			break;
		}
		case 1:
		{
			Polynomial deriv = deriv_Isc1 + insulin_rate_1;
			ode.assignDerivative("Isc1", deriv);
			break;
		}
		case 2:
		{
			Polynomial deriv = deriv_Isc1 + insulin_rate_2;
			ode.assignDerivative("Isc1", deriv);
			break;
		}
		case 3:
		{
			Polynomial deriv = deriv_Isc1 + insulin_rate_3;
			ode.assignDerivative("Isc1", deriv);
			break;
		}
		case 4:
		{
			Polynomial deriv = deriv_Isc1 + insulin_rate_4;
			ode.assignDerivative("Isc1", deriv);
			break;
		}
		}

		if(current_config.steps <= 30/stepsize)				// t <= 30
		{
			Polynomial deriv = deriv_Gp + meal_coeff_1;
			ode.assignDerivative("Gp", deriv);
		}
		else if(current_config.steps <= 80/stepsize)		// 30 <= t <= 80
		{
			Polynomial deriv = deriv_Gp + meal_coeff_2;
			ode.assignDerivative("Gp", deriv);
		}
		else if(current_config.steps <= 360/stepsize)		// 80 <= t <= 360
		{
			Polynomial deriv = deriv_Gp + meal_coeff_3;
			ode.assignDerivative("Gp", deriv);
		}
		else if(current_config.steps <= 400/stepsize)		// 360 <= t <= 400
		{
			Polynomial deriv = deriv_Gp + meal_coeff_4;
			ode.assignDerivative("Gp", deriv);
		}
		else if(current_config.steps <= 500/stepsize)		// 400 <= t <= 500
		{
			Polynomial deriv = deriv_Gp + meal_coeff_5;
			ode.assignDerivative("Gp", deriv);
		}
		else												// 500 <= t <= 720
		{
			Polynomial deriv = deriv_Gp + meal_coeff_6;
			ode.assignDerivative("Gp", deriv);
		}

		Flowpipe fpnext;
		bool bvalid = simulate(reachableSet, fpnext, current_config.flowpipe, ode, rs, N);

		if(!bvalid)
		{
			break;
		}

		std::vector<TaylorModel> tmFlowpipe, contracted_tmFlowpipe;
		std::vector<Interval> domain, contracted_domain;

		fpnext.toTaylorModel(tmFlowpipe, domain, rs);
		contracted_domain = domain;
		contracted_tmFlowpipe = tmFlowpipe;

		// check the jump condition
		switch(current_config.modeID)
		{
		case 0:
		{
			// we first check the mode invariant G <= 80
			int res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, inv_mode_0, rs);

			if(res >= 0)		// branch in the mode
			{
				// flowpipe intersects the invariant
				int steps = current_config.steps + N;

				if(steps < total_steps)
				{
					if(res == 0)		// whole flowpipe is in the invariant
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 0)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 0.\n", steps*stepsize);
								iter->flowpipe.merge(tmFlowpipe, domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							HybridConfig new_config(fpnext, 0, steps, current_config.jumps);
							queue.push_back(new_config);
						}
					}
					else
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 0)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 0.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_current_mode(contracted_tmFlowpipe, contracted_domain);
							HybridConfig new_config(fp_current_mode, 0, steps, current_config.jumps);
							queue.push_back(new_config);
						}
					}
				}
			}

			if(res != 0)		// branch to mode 1
			{
				contracted_domain = domain;
				contracted_tmFlowpipe = tmFlowpipe;

				res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, mode_0_to_1, rs);

				if(res != -1)
				{
					int steps = current_config.steps + N;
					int jmps = current_config.jumps + 1;

					if(steps < total_steps && jmps <= max_jumps)
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 1)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 1.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_next_mode(contracted_tmFlowpipe, contracted_domain);		// go to mode 1
							HybridConfig new_config(fp_next_mode, 1, steps, jmps);
							queue.push_back(new_config);
						}
					}
				}
			}

			break;
		}
		case 1:
		{
			// we first check the mode invariant 75 <= G <= 120
			int res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, inv_mode_1, rs);

			if(res >= 0)		// branch in the mode
			{
				// flowpipe intersects the invariant
				int steps = current_config.steps + N;

				if(steps < total_steps)
				{
					if(res == 0)		// whole flowpipe is in the invariant
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 1)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 1.\n", steps*stepsize);
								iter->flowpipe.merge(tmFlowpipe, domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							HybridConfig new_config(fpnext, 1, steps, current_config.jumps);
							queue.push_back(new_config);
						}
					}
					else
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 1)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 1.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_current_mode(contracted_tmFlowpipe, contracted_domain);
							HybridConfig new_config(fp_current_mode, 1, steps, current_config.jumps);
							queue.push_back(new_config);
						}
					}
				}
			}

			if(res != 0)
			{
				// check the jump condition for mode 1 -> mode 2
				contracted_domain = domain;
				contracted_tmFlowpipe = tmFlowpipe;
				res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, mode_1_to_2, rs);

				if(res != -1)
				{
					int steps = current_config.steps + N;
					int jmps = current_config.jumps + 1;

					if(steps < total_steps && jmps <= max_jumps)
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 2)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 2.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_next_mode(contracted_tmFlowpipe, contracted_domain);		// go to mode 2
							HybridConfig new_config(fp_next_mode, 2, steps, jmps);
							queue.push_back(new_config);
						}
					}
				}


				// check the jump condition for mode 1 -> mode 0
				contracted_domain = domain;
				contracted_tmFlowpipe = tmFlowpipe;
				res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, mode_1_to_0, rs);

				if(res != -1)
				{
					int steps = current_config.steps + N;
					int jmps = current_config.jumps + 1;

					if(steps < total_steps && jmps <= max_jumps)
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 0)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 0.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_next_mode(contracted_tmFlowpipe, contracted_domain);		// go to mode 0
							HybridConfig new_config(fp_next_mode, 0, steps, jmps);
							queue.push_back(new_config);
						}
					}
				}
			}

			break;
		}
		case 2:
		{
			// we first check the mode invariant 115 <= G <= 180
			int res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, inv_mode_2, rs);

			if(res >= 0)		// branch in the mode
			{
				// flowpipe intersects the invariant
				int steps = current_config.steps + N;

				if(steps < total_steps)
				{
					if(res == 0)		// whole flowpipe is in the invariant
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 2)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 2.\n", steps*stepsize);
								iter->flowpipe.merge(tmFlowpipe, domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							HybridConfig new_config(fpnext, 2, steps, current_config.jumps);
							queue.push_back(new_config);
						}
					}
					else
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 2)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 2.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_current_mode(contracted_tmFlowpipe, contracted_domain);
							HybridConfig new_config(fp_current_mode, 2, steps, current_config.jumps);
							queue.push_back(new_config);
						}
					}
				}
			}

			if(res != 0)
			{
				// check the jump condition for mode 2 -> mode 3
				contracted_domain = domain;
				contracted_tmFlowpipe = tmFlowpipe;
				res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, mode_2_to_3, rs);

				if(res != -1)
				{
					int steps = current_config.steps + N;
					int jmps = current_config.jumps + 1;

					if(steps < total_steps && jmps <= max_jumps)
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 3)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 3.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_next_mode(contracted_tmFlowpipe, contracted_domain);		// go to mode 3
							HybridConfig new_config(fp_next_mode, 3, steps, jmps);
							queue.push_back(new_config);
						}
					}
				}


				// check the jump condition for mode 2 -> mode 1
				contracted_domain = domain;
				contracted_tmFlowpipe = tmFlowpipe;
				res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, mode_2_to_1, rs);

				if(res != -1)
				{
					int steps = current_config.steps + N;
					int jmps = current_config.jumps + 1;

					if(steps < total_steps && jmps <= max_jumps)
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 1)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 1.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_next_mode(contracted_tmFlowpipe, contracted_domain);		// go to mode 1
							HybridConfig new_config(fp_next_mode, 1, steps, jmps);
							queue.push_back(new_config);
						}
					}
				}
			}

			break;
		}
		case 3:
		{
			// we first check the mode invariant 175 <= G <= 300
			int res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, inv_mode_3, rs);

			if(res >= 0)		// branch in the mode
			{
				// flowpipe intersects the invariant
				int steps = current_config.steps + N;

				if(steps < total_steps)
				{
					if(res == 0)		// whole flowpipe is in the invariant
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 3)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 3.\n", steps*stepsize);
								iter->flowpipe.merge(tmFlowpipe, domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							HybridConfig new_config(fpnext, 3, steps, current_config.jumps);
							queue.push_back(new_config);
						}
					}
					else
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 3)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 3.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_current_mode(contracted_tmFlowpipe, contracted_domain);
							HybridConfig new_config(fp_current_mode, 3, steps, current_config.jumps);
							queue.push_back(new_config);
						}
					}
				}
			}

			if(res != 0)
			{
				// check the jump condition for mode 3 -> mode 4
				contracted_domain = domain;
				contracted_tmFlowpipe = tmFlowpipe;
				res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, mode_3_to_4, rs);

				if(res != -1)
				{
					int steps = current_config.steps + N;
					int jmps = current_config.jumps + 1;

					if(steps < total_steps && jmps <= max_jumps)
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 4)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 4.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_next_mode(contracted_tmFlowpipe, contracted_domain);		// go to mode 4
							HybridConfig new_config(fp_next_mode, 4, steps, jmps);
							queue.push_back(new_config);
						}
					}
				}


				// check the jump condition for mode 3 -> mode 2
				contracted_domain = domain;
				contracted_tmFlowpipe = tmFlowpipe;
				res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, mode_3_to_2, rs);

				if(res != -1)
				{
					int steps = current_config.steps + N;
					int jmps = current_config.jumps + 1;

					if(steps < total_steps && jmps <= max_jumps)
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 2)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 2.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_next_mode(contracted_tmFlowpipe, contracted_domain);		// go to mode 2
							HybridConfig new_config(fp_next_mode, 2, steps, jmps);
							queue.push_back(new_config);
						}
					}
				}
			}

			break;
		}
		case 4:
		{
			// we first check the mode invariant G >= 295
			int res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, inv_mode_4, rs);

			if(res >= 0)		// branch in the mode
			{
				// flowpipe intersects the invariant
				int steps = current_config.steps + N;

				if(steps < total_steps)
				{
					if(res == 0)		// whole flowpipe is in the invariant
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 4)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 4.\n", steps*stepsize);
								iter->flowpipe.merge(tmFlowpipe, domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							HybridConfig new_config(fpnext, 4, steps, current_config.jumps);
							queue.push_back(new_config);
						}
					}
					else
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 4)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 4.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_current_mode(contracted_tmFlowpipe, contracted_domain);
							HybridConfig new_config(fp_current_mode, 4, steps, current_config.jumps);
							queue.push_back(new_config);
						}
					}
				}
			}

			if(res != 0)
			{
				// check the jump condition for mode 4 -> mode 3
				contracted_domain = domain;
				contracted_tmFlowpipe = tmFlowpipe;
				res = contract_flowpipe(contracted_tmFlowpipe, contracted_domain, mode_4_to_3, rs);

				if(res != -1)
				{
					int steps = current_config.steps + N;
					int jmps = current_config.jumps + 1;

					if(steps < total_steps && jmps <= max_jumps)
					{
						// try to merge the current flowpipe to an existing one
						std::list<HybridConfig>::iterator iter = queue.begin();
						for(; iter != queue.end(); ++iter)
						{
							if(iter->steps == steps && iter->modeID == 3)
							{
								// merge the flowpipes
//								printf("Time = %f: Merged to a flowpipe in Mode 3.\n", steps*stepsize);
								iter->flowpipe.merge(contracted_tmFlowpipe, contracted_domain, rs);
								break;
							}
						}

						if(iter == queue.end())		// flowpipe is not merged
						{
							Flowpipe fp_next_mode(contracted_tmFlowpipe, contracted_domain);		// go to mode 3
							HybridConfig new_config(fp_next_mode, 3, steps, jmps);
							queue.push_back(new_config);
						}
					}
				}
			}

			break;
		}
		}
	}

	end = clock();
	printf("time cost: %lf\n", (double)(end - begin) / CLOCKS_PER_SEC);

	FILE *fp = fopen("insulin_no_noise_Strategy_2.m", "w");

	printf("Generating the plot file...\n");
	plot_2D_box_gnuplot(fp, "insulin", reachableSet, "t", "Gs", stateVars, cutoff);
	printf("Done.\n");

	fclose(fp);

	return 0;
}


